/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/
/* $XConsortium: mipolyconZ.c,v 1.4 93/12/09 16:14:01 takamori Exp $ */
#include "gcstruct.h"
#include "pixmap.h"
#include "miscanfillZ.h"
#include "ddpex.h"
#include "ddpex3.h"

static int getPolyYBounds();

/*
 *     convexpoly.c
 *
 *     Written by Brian Kelleher; Dec. 1985.
 *
 *     Fill a convex polygon.  If the given polygon
 *     is not convex, then the result is undefined.
 *     The algorithm is to order the edges from smallest
 *     y to largest by partitioning the array into a left
 *     edge list and a right edge list.  The algorithm used
 *     to traverse each edge is an extension of Bresenham's
 *     line algorithm with y as the major axis.
 *     For a derivation of the algorithm, see the author of
 *     this code.
 */
Bool
miZFillConvexPoly(zbs_ptr, dst, pgc, count, ptsIn)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr dst;
    GCPtr	pgc;
    int		count;                /* number of points           */
    ddCoord3DS  *ptsIn;               /* the points                 */
{
    register int xl, xr;        /* x vals of left and right edges   */
    register int zl, zr;        /* z vals of left and right edges   */
    register int dl, dr;        /* decision variables               */
    register int dzl, dzr;      /* decision variables for z increment */
    register int ml, m1l;       /* left edge slope and slope+1      */
    register int mzl, mz1l;     /* left edge z-slope and z-slope+1  */
    int mr, m1r;                /* right edge slope and slope+1     */
    int mzr, mz1r;              /* right edge z-slope and z-slope+1 */
    int incr1l, incr2l;         /* left edge error increments       */
    int incrz1l, incrz2l;       /* left edge z error increments     */
    int incr1r, incr2r;         /* right edge error increments      */
    int incrz1r, incrz2r;       /* right edge z error increments    */
    int dy;                     /* delta y                          */
    int y;                      /* current scanline                 */
    int left, right;            /* indices to first endpoints       */
    int i;                      /* loop counter                     */
    int nextleft, nextright;    /* indices to second endpoints      */
    ddCoord3DS *ptsOut, *FirstPoint; /* output buffer               */
    int *width, *FirstWidth;    /* output buffer                    */
    int *depth, *FirstDepth;    /* output buffer                    */
    int imin;                   /* index of smallest vertex (in y)  */
    int ymin;                   /* y-extents of polygon             */
    int ymax;

    /*
     *  find leftx, bottomy, rightx, topy, and the index
     *  of bottomy. Also translate the points.
     */
    imin = getPolyYBounds(ptsIn, count, &ymin, &ymax);

    dy = ymax - ymin + 1;
    if ((count < 3) || (dy < 0))
	return(TRUE);
    ptsOut = FirstPoint = (ddCoord3DS *)ALLOCATE_LOCAL(sizeof(ddCoord3DS)*dy);
    width = FirstWidth = (int *)ALLOCATE_LOCAL(sizeof(int) * dy);
    depth = FirstDepth = (int *)ALLOCATE_LOCAL(sizeof(int) * dy);
    if(!FirstPoint || !FirstWidth || !FirstDepth)
    {
	if (FirstDepth) DEALLOCATE_LOCAL(FirstDepth);
	if (FirstWidth) DEALLOCATE_LOCAL(FirstWidth);
	if (FirstPoint) DEALLOCATE_LOCAL(FirstPoint);
	return(FALSE);
    }

    nextleft = nextright = imin;
    y = ptsIn[nextleft].y;

    /*
     *  loop through all edges of the polygon
     */
    do {
        /*
         *  add a left edge if we need to
         */
        if (ptsIn[nextleft].y == y) {
            left = nextleft;

            /*
             *  find the next edge, considering the end
             *  conditions of the array.
             */
            nextleft++;
            if (nextleft >= count)
                nextleft = 0;

            /*
             *  now compute all of the random information
             *  needed to run the iterative algorithm.
             */
            ZBRESINITPGON(ptsIn[nextleft].y-ptsIn[left].y,
                         ptsIn[left].x,ptsIn[nextleft].x,
                         ptsIn[left].z,ptsIn[nextleft].z,
                         xl, zl, dl, dzl, ml, m1l, mzl, mz1l,
			 incr1l, incr2l, incrz1l, incrz2l);
        }

        /*
         *  add a right edge if we need to
         */
        if (ptsIn[nextright].y == y) {
            right = nextright;

            /*
             *  find the next edge, considering the end
             *  conditions of the array.
             */
            nextright--;
            if (nextright < 0)
                nextright = count-1;

            /*
             *  now compute all of the random information
             *  needed to run the iterative algorithm.
             */
            ZBRESINITPGON(ptsIn[nextright].y-ptsIn[right].y,
                         ptsIn[right].x,ptsIn[nextright].x,
                         ptsIn[right].z,ptsIn[nextright].z,
                         xr, zr, dr, dzr, mr, m1r, mzr, mz1r,
			 incr1r, incr2r, incrz1r, incrz2r);
        }

        /*
         *  generate scans to fill while we still have
         *  a right edge as well as a left edge.
         */
        i = min(ptsIn[nextleft].y, ptsIn[nextright].y) - y;
	/* in case we're called with non-convex polygon */
	if(i < 0)
        {
	    DEALLOCATE_LOCAL(FirstDepth);
	    DEALLOCATE_LOCAL(FirstWidth);
	    DEALLOCATE_LOCAL(FirstPoint);
	    return(TRUE);
	}
        while (i-- > 0) 
        {
            ptsOut->y = y;

            /*
             *  reverse the edges if necessary
             */
            if (xl < xr) 
            {
                *(depth++) = zr - zl;
                *(width++) = xr - xl;
                ptsOut->x = xl;
                (ptsOut++)->z = zl;
            }
            else
            {
                *(depth++) = zl - zr;
                *(width++) = xl - xr;
                ptsOut->x = xr;
                (ptsOut++)->z = zr;
            }
            y++;

            /* increment down the edges */
            ZBRESINCRPGON(dl, xl, ml, m1l, incr1l, incr2l);
            ZBRESINCRPGON(dzl, zl, mzl, mz1l, incrz1l, incrz2l);
            ZBRESINCRPGON(dr, xr, mr, m1r, incr1r, incr2r);
            ZBRESINCRPGON(dzr, zr, mzr, mz1r, incrz1r, incrz2r);
        }
    }  while (y != ymax);

    /*
     * Finally, fill the <remaining> spans
     */
    ZBuffer_FillSpans(zbs_ptr, dst, pgc,
		      ptsOut-FirstPoint, FirstPoint, FirstWidth, FirstDepth);
    DEALLOCATE_LOCAL(FirstDepth);
    DEALLOCATE_LOCAL(FirstWidth);
    DEALLOCATE_LOCAL(FirstPoint);
    return(TRUE);
}

/*
 *     Find the index of the point with the smallest y.
 */
static
int
getPolyYBounds(pts, n, by, ty)
    ddCoord3DS *pts;
    int n;
    int *by, *ty;
{
    register ddCoord3DS *ptMin;
    int ymin, ymax;
    ddCoord3DS *ptsStart = pts;

    ptMin = pts;
    ymin = ymax = (pts++)->y;

    while (--n > 0) {
        if (pts->y < ymin)
	{
            ptMin = pts;
            ymin = pts->y;
        }
	if(pts->y > ymax)
            ymax = pts->y;

        pts++;
    }

    *by = ymin;
    *ty = ymax;
    return(ptMin-ptsStart);
}
